from abc import ABC, abstractmethod
from functools import partial
from typing import Awaitable, override, Union, Any, Literal, Callable

from greyhorse.maybe import Maybe, Just, Nothing
from greyhorse.utils.types import TypeWrapper


class Operator[T](TypeWrapper[T], ABC):
    @abstractmethod
    def accept(self, instance: T) -> Union[bool, Awaitable[bool]]:
        ...

    @abstractmethod
    def revoke(self) -> Maybe[T] | Awaitable[Maybe[T]]:
        ...


class AssignOperator[T](Operator[T]):
    def __init__(self, getter: Callable[[], Maybe[T]], setter: Callable[[Maybe[T]], Any]):
        super().__init__()
        self._getter = getter
        self._setter = setter

    def _set(self, value: T) -> Literal[True]:
        self._setter(Just(value))
        return True

    def _clear(self, value: T) -> T:
        self._setter(Nothing)
        return value

    @override
    def accept(self, instance: T) -> bool:
        return self._getter().map_or_else(partial(self._set, instance), lambda _: False)

    @override
    def revoke(self) -> Maybe[T]:
        return self._getter().map(self._clear)


class AttrOperator[T](AssignOperator[T]):
    def __init__(self, instance: object, attr: str):
        if hasattr(instance, attr):
            super().__init__(partial(getattr, instance, attr), partial(setattr, instance, attr))
        else:
            raise ValueError(f'Instance {repr(instance)} has no attribute "{attr}"')



from collections import defaultdict
from typing import override, Callable

from greyhorse.app.abc.collectors import Collector, MutCollector
from greyhorse.app.abc.selectors import ListSelector
from greyhorse.maybe import Nothing, Maybe, Just


class DictRegistry[K, T](Collector[K, T], ListSelector[K, T]):
    __slots__ = ('_storage',)

    def __init__(self):
        self._storage: dict[K, list[T]] = defaultdict(list)

    def __len__(self):
        return len(self._storage)

    @override
    def add(self, key: K, instance: T) -> bool:
        if key in self._storage:
            return instance is self._storage[key]
        self._storage[key].append(instance)
        return True

    @override
    def has(self, key: K) -> bool:
        if key not in self._storage:
            return False
        return len(self._storage[key]) > 0

    @override
    def get(self, key: K) -> Maybe[T]:
        if key not in self._storage:
            return Nothing

        values = self._storage[key]
        return Just(values[0]) if len(values) > 0 else Nothing

    @override
    def items(self, filter_fn: Callable[[K], bool] | None = None) -> list[tuple[K, T]]:
        result = []

        for k, values in self._storage.items():
            for v in values:
                if not filter_fn or filter_fn(k):
                    result.append((k, v))

        return result


class MutDictRegistry[K, T](MutCollector[K, T], DictRegistry[K, T]):
    @override
    def remove(self, key: K, instance: T) -> bool:
        if key not in self._storage:
            return False

        values = self._storage[key]
        if instance not in values:
            return False

        values.remove(instance)

        if len(values) == 0:
            del self._storage[key]

        return True


class ScopedDictRegistry[K, T](DictRegistry[K, T]):
    def __init__(
        self, factory: Callable[[], DictRegistry[K, T]], scope_func: Callable[[], str],
    ):
        super().__init__()
        self._scope_func = scope_func
        self._storage: dict[str, DictRegistry[K, T]] = defaultdict(factory)

    def _get_registry(self) -> DictRegistry[K, T]:
        key = self._scope_func()
        return self._storage[key]

    def __len__(self):
        registry = self._get_registry()
        return registry.__len__()

    @override
    def add(self, key: K, instance: T) -> bool:
        registry = self._get_registry()
        return registry.add(key, instance)

    @override
    def has(self, key: K) -> bool:
        registry = self._get_registry()
        return registry.has(key)

    @override
    def get(self, key: K) -> Maybe[T]:
        registry = self._get_registry()
        return registry.get(key)

    @override
    def items(self, filter_fn: Callable[[K], bool] | None = None) -> list[tuple[K, T]]:
        registry = self._get_registry()
        return registry.items(filter_fn)


class ScopedMutDictRegistry[K, T](MutDictRegistry[K, T]):
    def __init__(
        self, factory: Callable[[], MutDictRegistry[K, T]], scope_func: Callable[[], str],
    ):
        super().__init__()
        self._scope_func = scope_func
        self._storage: dict[str, MutDictRegistry[K, T]] = defaultdict(factory)

    def _get_registry(self) -> MutDictRegistry[K, T]:
        key = self._scope_func()
        return self._storage[key]

    def __len__(self):
        registry = self._get_registry()
        return registry.__len__()

    @override
    def add(self, key: K, instance: T) -> bool:
        registry = self._get_registry()
        return registry.add(key, instance)

    @override
    def has(self, key: K) -> bool:
        registry = self._get_registry()
        return registry.has(key)

    @override
    def get(self, key: K) -> Maybe[T]:
        registry = self._get_registry()
        return registry.get(key)

    @override
    def items(self, filter_fn: Callable[[K], bool] | None = None) -> list[tuple[K, T]]:
        registry = self._get_registry()
        return registry.items(filter_fn)

    @override
    def remove(self, key: K, instance: T) -> bool:
        registry = self._get_registry()
        return registry.remove(key, instance)



from typing import TYPE_CHECKING

from greyhorse.app.abc.collectors import Collector, MutCollector
from greyhorse.app.abc.controllers import Controller, ControllerError, ControllerFactoryFn
from greyhorse.app.abc.operators import Operator
from greyhorse.app.abc.providers import Provider
from greyhorse.app.abc.selectors import ListSelector
from greyhorse.app.abc.services import Service, ServiceFactoryFn, ServiceError
from greyhorse.app.registries import MutDictRegistry
from greyhorse.app.schemas.components import ComponentConf, ModuleComponentConf
from greyhorse.app.schemas.elements import CtrlConf, SvcConf
from greyhorse.error import Error, ErrorCase
from greyhorse.logging import logger
from greyhorse.maybe import Maybe
from greyhorse.result import Result, Ok, Err
from greyhorse.utils.injectors import ParamsInjector

if TYPE_CHECKING:
    from .module import Module


class ComponentError(Error):
    namespace = 'greyhorse.app.component'

    Ctrl = ErrorCase(
        msg='{path}: Controller error in component: "{name}", details: "{details}"',
        path=str, name=str, details=str,
    )
    Service = ErrorCase(
        msg='{path}: Service error in component: "{name}", details: "{details}"',
        path=str, name=str, details=str,
    )
    Module = ErrorCase(
        msg='{path}: Submodule error in component: "{name}", details: "{details}"',
        path=str, name=str, details=str,
    )


class BasicComponent:
    def __init__(self, name: str, path: str):
        self._name = name
        self._path = path

    @property
    def name(self) -> str:
        return self._name

    @property
    def path(self) -> str:
        return self._path


class Component(BasicComponent):
    def __init__(
        self, name: str, path: str, conf: ComponentConf,
        operator_reg: MutDictRegistry[type, Operator],
    ):
        super().__init__(name, path)
        self._conf = conf

        self._controllers: list[Controller] = []
        self._services: dict[type[Service], Service] = {}

        self._imported_providers = MutDictRegistry[type[Provider], Provider]()
        self._private_providers = MutDictRegistry[type[Provider], Provider]()
        self._operator_reg = operator_reg

    def get_service[T: Service](self, type_: type[T]) -> Maybe[T]:
        return Maybe(self._services.get(type_))

    def setup(
        self, prov_selector: ListSelector[type[Provider], Provider],
        prov_collector: Collector[type[Provider], Provider],
    ) -> Result[None, ComponentError]:
        injector = ParamsInjector()

        logger.info(
            '{path}: Component "{name}" setup'
            .format(path=self._path, name=self.name)
        )

        for prov_conf in self._conf.provider_grants:
            for prov_type in prov_conf.types:
                for _, prov in prov_selector.items(lambda t: issubclass(t, prov_type)):
                    self._imported_providers.add(prov_type, prov)

        for prov_type, provider in self._imported_providers.items():
            injector.add_type_provider(prov_type, provider)

        if not (res := self._create_controllers(injector)):
            return res

        self._controllers = res.unwrap()

        for ctrl in self._controllers:
            if not (res := ctrl.setup(self._imported_providers, self._operator_reg).map_err(
                lambda e: ComponentError.Ctrl(path=self._path, name=self.name, details=e.message)
            )):
                return res

        for res_type, operator in self._operator_reg.items():
            injector.add_type_provider(Operator[res_type], operator)

        if not (res := self._create_services(injector)):
            return res

        self._services = {type(s): s for s in res.unwrap()}

        for svc in self._services.values():
            if not (res := svc.setup(self._operator_reg, self._private_providers).map_err(
                lambda e: ComponentError.Service(path=self._path, name=self.name, details=e.message)
            )):
                return res

        for prov_conf in self._conf.provider_imports:
            for prov_type in prov_conf.types:
                for _, prov in self._private_providers.items(lambda t: issubclass(t, prov_type)):
                    prov_collector.add(prov_type, prov)

        logger.info(
            '{path}: Component "{name}" setup successful'
            .format(path=self._path, name=self.name)
        )

        return Ok(None)

    def teardown(
        self, prov_selector: ListSelector[type[Provider], Provider],
        prov_collector: MutCollector[type[Provider], Provider],
    ) -> Result[None, ComponentError]:

        logger.info(
            '{path}: Component "{name}" teardown'
            .format(path=self._path, name=self.name)
        )

        for prov_conf in reversed(self._conf.provider_imports):
            for prov_type in prov_conf.types:
                for _, prov in self._private_providers.items(lambda t: issubclass(t, prov_type)):
                    prov_collector.remove(prov_type, prov)

        for svc in reversed(self._services.values()):
            if not (res := svc.teardown(self._operator_reg, self._private_providers).map_err(
                lambda e: ComponentError.Service(path=self._path, name=self.name, details=e.message)
            )):
                return res

        self._services = {}

        for ctrl in reversed(self._controllers):
            if not (res := ctrl.teardown(self._imported_providers, self._operator_reg).map_err(
                lambda e: ComponentError.Ctrl(path=self._path, name=self.name, details=e.message)
            )):
                return res

        self._controllers = []

        for prov_conf in reversed(self._conf.provider_grants):
            for prov_type in prov_conf.types:
                for _, prov in prov_selector.items(lambda t: issubclass(t, prov_type)):
                    self._imported_providers.remove(prov_type, prov)

        logger.info(
            '{path}: Component "{name}" teardown successful'
            .format(path=self._path, name=self.name)
        )

        return Ok(None)

    def _create_controller(
        self, conf: CtrlConf, factory: ControllerFactoryFn,
        injector: ParamsInjector,
    ) -> Result[Controller, ControllerError]:
        logger.info(
            '{path}: Component controller "{name}" create'
            .format(path=self._path, name=conf.name)
        )

        values = conf.args.copy()
        values['name'] = conf.name
        values['operators'] = conf.operators
        injected_args = injector(factory, values=values)

        try:
            if not (res := factory(*injected_args.args, **injected_args.kwargs)):
                return res

        except Exception as e:
            error = ControllerError.Factory(details=str(e))
            logger.error(error.message)
            return error.to_result()

        logger.info(
            '{path}: Component controller "{name}" created successfully'
            .format(path=self._path, name=conf.name)
        )

        if isinstance(res, Controller):
            return Ok(res)

        return res

    def _create_service(
        self, conf: SvcConf, factory: ServiceFactoryFn,
        injector: ParamsInjector,
    ) -> Result[Service, ServiceError]:
        logger.info(
            '{path}: Component service "{name}" create'
            .format(path=self._path, name=conf.name)
        )

        values = conf.args.copy()
        values['name'] = conf.name
        values['providers'] = conf.providers
        injected_args = injector(factory, values=values)

        try:
            if not (res := factory(*injected_args.args, **injected_args.kwargs)):
                return res

        except Exception as e:
            error = ServiceError.Factory(details=str(e))
            logger.error(error.message)
            return error.to_result()

        logger.info(
            '{path}: Component service "{name}" created successfully'
            .format(path=self._path, name=conf.name)
        )

        if isinstance(res, Service):
            return Ok(res)

        return res

    def _create_controllers(
        self, injector: ParamsInjector,
    ) -> Result[list[Controller], ComponentError]:
        result = []

        for conf in self._conf.controllers:
            factory = self._conf.controller_factories.get(conf.type_, conf.type_)

            match self._create_controller(conf, factory, injector):
                case Ok(ctrl):
                    result.append(ctrl)

                case Err(e):
                    error = ComponentError.Ctrl(path=self._path, name=self.name, details=e.message)
                    logger.error(error.message)
                    return error.to_result()

        return Ok(result)

    def _create_services(
        self, injector: ParamsInjector,
    ) -> Result[list[Service], ComponentError]:
        result = []

        for conf in self._conf.services:
            factory = self._conf.service_factories.get(conf.type_, conf.type_)

            match self._create_service(conf, factory, injector):
                case Ok(svc):
                    result.append(svc)

                case Err(e):
                    error = ComponentError.Service(path=self._path, name=self.name, details=e.message)
                    logger.error(error.message)
                    return error.to_result()

        return Ok(result)


class ModuleComponent(BasicComponent):
    def __init__(
        self, name: str, path: str,
        conf: ModuleComponentConf, module: 'Module',
    ):
        super().__init__(name, path)
        self._conf = conf
        self._module = module

    def setup(
        self, prov_selector: ListSelector[type[Provider], Provider],
        prov_collector: Collector[type[Provider], Provider],
    ) -> Result[None, ComponentError]:
        for prov_conf in self._conf.provider_grants:
            for prov_type in prov_conf.types:
                for _, prov in prov_selector.items(lambda t: issubclass(t, prov_type)):
                    self._module.add_provider(prov_type, prov)

        if not (res := self._module.setup().map_err(
            lambda e: ComponentError.Module(path=self._path, name=self.name, details=e.message)
        )):
            return res

        for prov_conf in self._conf.provider_imports:
            for prov_type in prov_conf.types:
                if prov := self._module.get_provider(prov_type).unwrap_or_none():
                    prov_collector.add(prov_type, prov)

        return Ok(None)

    def teardown(
        self, prov_selector: ListSelector[type[Provider], Provider],
        prov_collector: MutCollector[type[Provider], Provider],
    ) -> Result[None, ComponentError]:
        for prov_conf in reversed(self._conf.provider_imports):
            for prov_type in prov_conf.types:
                if prov := self._module.get_provider(prov_type).unwrap_or_none():
                    prov_collector.remove(prov_type, prov)

        if not (res := self._module.teardown().map_err(
            lambda e: ComponentError.Module(path=self._path, name=self.name, details=e.message)
        )):
            return res

        for prov_conf in reversed(self._conf.provider_grants):
            for prov_type in prov_conf.types:
                for _, prov in prov_selector.items(lambda t: issubclass(t, prov_type)):
                    self._module.remove_provider(prov_type, prov)

        return Ok(None)



import asyncio
import inspect
from asyncio import Future, get_running_loop, iscoroutine, iscoroutinefunction, run as run_main
from functools import partial
from typing import Callable


def is_awaitable(f):
    while isinstance(f, partial):
        f = f.func
    return iscoroutinefunction(f) or inspect.isawaitable(f)


def is_like_sync_context_manager(instance) -> bool:
    if callable(instance):
        if inspect.isgeneratorfunction(inspect.unwrap(instance)):
            instance = instance()
    return hasattr(instance, '__enter__') and hasattr(instance, '__exit__')


def is_like_async_context_manager(instance) -> bool:
    if callable(instance):
        if inspect.isasyncgenfunction(inspect.unwrap(instance)):
            instance = instance()
    return hasattr(instance, '__aenter__') and hasattr(instance, '__aexit__')


def get_asyncio_loop():
    try:
        return get_running_loop()
    except RuntimeError:
        return None


def invoke_sync[T, **P](func: Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
    if is_awaitable(func):
        if loop := get_asyncio_loop():
            return loop.run_until_complete(func(*args, **kwargs))
        else:
            return run_main(func(*args, **kwargs))
    elif callable(func):
        return func(*args, **kwargs)
    else:
        return func


async def invoke_async[T, **P](
    func: Callable[P, T], *args: P.args, to_thread: bool = False, **kwargs: P.kwargs,
) -> T:
    if is_awaitable(func):
        if iscoroutine(func):
            return await func
        return await func(*args, **kwargs)
    elif callable(func):
        if to_thread:
            return await asyncio.to_thread(func, *args, **kwargs)
        else:
            return func(*args, **kwargs)
    else:
        future = Future()
        future.set_result(func)
        return future


def caller_path(depth: int) -> list[str]:
    import inspect

    frame = inspect.currentframe()

    for _ in range(depth, 0, -1):
        frame = frame.f_back

    path = inspect.getmodule(frame).__name__.split('.')
    return path



from abc import abstractmethod
from typing import Any, Generic, Mapping, Self, Sequence, TYPE_CHECKING, TypeVar

from .base import IdType
from .model import Model, UpdateSchemaType

if TYPE_CHECKING:
    from ..repositories.filterable import FilterableRepository


FilterType = TypeVar('FilterType')
SortingType = TypeVar('SortingType')


class FilterableModel(Model[IdType], Generic[IdType, FilterType, SortingType]):
    _repo: 'FilterableRepository[IdType, Self, FilterType, SortingType]'

    @classmethod
    def bind(cls, repository: 'FilterableRepository[IdType, Self, FilterType, SortingType]'):
        cls._repo = repository

    @classmethod
    @abstractmethod
    def get_columns(cls) -> Mapping[str, Any]:
        ...

    @classmethod
    @abstractmethod
    def get_relationships(cls) -> Mapping[str, Any]:
        ...

    #
    # Retrieve operations
    #

    @classmethod
    async def list(
        cls, filters: FilterType | None = None,
        sorting: SortingType | None = None,
        skip: int = 0, limit: int = 0, **kwargs,
    ) -> Sequence[Self]:
        return await cls._repo.list(filters, sorting, skip, limit, **kwargs)

    @classmethod
    async def count(cls, filters: FilterType | None = None, **kwargs) -> int:
        return await cls._repo.count(filters, **kwargs)

    @classmethod
    async def exists_by(cls, filters: FilterType, **kwargs) -> bool:
        return await cls._repo.exists_by(filters, **kwargs)

    #
    # Modification operations
    #

    @classmethod
    async def update_by(
        cls, filters: FilterType, data: UpdateSchemaType | Mapping[str, Any], **kwargs,
    ) -> int:
        if isinstance(data, Mapping):
            update_data = data
        else:
            update_data = data.dict(exclude_unset=True)

        if update_data:
            update_data = cls.prepare_for_update(update_data)

        return await cls._repo.update_by(filters, update_data, **kwargs)

    @classmethod
    async def delete_by(cls, filters: FilterType, **kwargs) -> int:
        return await cls._repo.delete_by(filters, **kwargs)



import asyncio
import inspect
import threading
from collections import defaultdict
from dataclasses import dataclass, field
from functools import partial
from typing import override, Any

from greyhorse.result import Result, Ok, Err
from ..abc.collectors import Collector, MutCollector
from ..abc.operators import Operator
from ..abc.providers import Provider
from ..abc.selectors import ListSelector
from ..abc.services import Service, ServiceWaiter, ServiceState, ServiceError, ProvisionError
from ..private.mappers import SyncResourceMapper, AsyncResourceMapper
from ...utils.invoke import invoke_sync, invoke_async


@dataclass(slots=True, frozen=True)
class _ProviderMember:
    class_name: str
    method_name: str
    resource_type: type
    provider_type: type[Provider]
    method: classmethod
    params: dict[str, inspect.Parameter] = field(default_factory=dict)


def _is_provider_member(member: Any) -> bool:
    return isinstance(member, _ProviderMember)


def provider(provider_type: type[Provider]):
    def decorator(func: classmethod):
        class_name, method_name = func.__qualname__.split('.')
        sig = inspect.signature(func)
        params = {k: v for k, v in sig.parameters.items() if k != 'self'}

        return _ProviderMember(
            class_name, method_name,
            provider_type.__wrapped_type__,
            provider_type, func, params,
        )

    return decorator


class SyncService(Service):
    def __init__(self, providers: list[type[Provider]] | None = None):
        super().__init__()
        self._public_providers = providers or []
        self._providers: dict[type, list[type[Provider]]] = defaultdict(list)
        self._provider_members: dict[type[Provider], _ProviderMember] = {}
        self._provided_resources: dict[type, dict[Operator, SyncResourceMapper]] = defaultdict(dict)
        self._provided_providers: dict[type[Provider], Provider] = {}
        self._waiter = threading.Event()
        self._init_provider_members()

    def _init_provider_members(self):
        provider_members = inspect.getmembers(self, _is_provider_member)

        for name, member in provider_members:  # type: str, _ProviderMember
            self._providers[member.resource_type].append(member.provider_type)
            self._provider_members[member.provider_type] = member
            setattr(self, name, partial(member.method, self))

    @override
    @property
    def waiter(self) -> ServiceWaiter:
        return ServiceWaiter.Sync(self._waiter)

    @override
    def setup(
        self, selector: ListSelector[type, Operator],
        collector: Collector[type[Provider], Provider],
    ) -> Result[ServiceState, ServiceError]:
        if self.state == ServiceState.Active:
            return Ok(self.state)

        for prov_type in self._public_providers:
            res_type = prov_type.__wrapped_type__

            for _, operator in selector.items(lambda t: issubclass(t, res_type)):
                res = self.setup_resource(prov_type, operator) \
                    .map_err(lambda e: ServiceError.AutoProvision(details=e.message))
                if not res:
                    return res

        for prov_type, member in self._provider_members.items():
            if len(member.params) > 0:
                continue

            # noinspection PyTypeChecker
            match invoke_sync(member.method, self):
                case Ok(prov) | (Provider() as prov):
                    if collector.add(prov_type, prov):
                        self._provided_providers[prov_type] = prov

                case Err(e):
                    return ServiceError.Collect(details=e.message).to_result()

        self._switch_to_active()
        return Ok(self.state)

    @override
    def teardown(
        self, selector: ListSelector[type, Operator],
        collector: MutCollector[type[Provider], Provider],
    ) -> Result[ServiceState, ServiceError]:
        if self.state == ServiceState.Idle:
            return Ok(self.state)

        for prov_type in reversed(self._public_providers):
            res_type = prov_type.__wrapped_type__

            for _, operator in selector.items(lambda t: issubclass(t, res_type)):
                res = self.teardown_resource(prov_type, operator) \
                    .map_err(lambda e: ServiceError.AutoProvision(details=e.message))
                if not res:
                    return res

        provs_to_remove = []

        for prov_type, prov in reversed(self._provided_providers.items()):
            if collector.remove(prov_type, prov):
                provs_to_remove.append(prov_type)
            else:
                # TODO logger warn
                pass

        for prov_type in provs_to_remove:
            del self._provided_providers[prov_type]

        self._switch_to_idle()
        return Ok(self.state)

    @override
    def can_provide(self, resource_type: type) -> bool:
        if resource_type not in self._providers:
            return False
        return len(self._providers[resource_type]) > 0

    @override
    def get_resource_types(self) -> list[type]:
        return list(self._providers.keys())

    @override
    def get_provider_types(self, resource_type: type) -> list[type[Provider]]:
        if resource_type not in self._providers:
            return []
        return self._providers[resource_type].copy()

    @override
    def setup_resource(
        self, prov_type: type[Provider], operator: Operator, *args, **kwargs,
    ) -> Result[None, ProvisionError]:
        if not (member := self._provider_members.get(prov_type)):
            return ProvisionError.NoSuchProvider(type_=prov_type.__name__).to_result()
        if operator in self._provided_resources[prov_type.__wrapped_type__]:
            return Ok(None)

        # noinspection PyTypeChecker
        match invoke_sync(member.method, self, *args, **kwargs):
            case Ok(prov) | (Provider() as prov):
                mapper = SyncResourceMapper[prov_type](prov, operator)

                if res := mapper.setup().map_err(lambda e: ProvisionError.Provision(details=e)):
                    self._provided_resources[prov_type.__wrapped_type__][operator] = mapper
                return res

            case Err(e):
                return e

    @override
    def teardown_resource(
        self, prov_type: type[Provider], operator: Operator,
    ) -> Result[None, ProvisionError]:
        if not (mappers := self._provided_resources.get(prov_type.__wrapped_type__)):
            return Ok(None)
        if not (mapper := mappers.get(operator)):
            return Ok(None)

        if res := mapper.teardown().map_err(lambda e: ProvisionError.Provision(details=e)):
            del mappers[operator]
        return res

    @override
    def _switch_to_idle(self):
        self._state = ServiceState.Idle

    @override
    def _switch_to_active(self, started: bool = False):
        if started:
            self._waiter.clear()
        else:
            self._waiter.set()

        self._state = ServiceState.Active(started=not self._waiter.is_set())


class AsyncService(Service):
    def __init__(self, providers: list[type[Provider]] | None = None):
        super().__init__()
        self._public_providers = providers or []
        self._providers: dict[type, list[type[Provider]]] = defaultdict(list)
        self._provider_members: dict[type[Provider], _ProviderMember] = {}
        self._provided_resources: dict[type, dict[Operator, AsyncResourceMapper]] = defaultdict(dict)
        self._provided_providers: dict[type[Provider], Provider] = {}
        self._waiter = asyncio.Event()
        self._init_provider_members()

    def _init_provider_members(self):
        provider_members = inspect.getmembers(self, _is_provider_member)

        for name, member in provider_members:  # type: str, _ProviderMember
            self._providers[member.resource_type].append(member.provider_type)
            self._provider_members[member.provider_type] = member
            setattr(self, name, partial(member.method, self))

    @override
    @property
    def waiter(self) -> ServiceWaiter:
        return ServiceWaiter.Async(self._waiter)

    @override
    async def setup(
        self, selector: ListSelector[type, Operator],
        collector: Collector[type[Provider], Provider],
    ) -> Result[ServiceState, ServiceError]:
        if self.state == ServiceState.Active:
            return Ok(self.state)

        for prov_type in self._public_providers:
            res_type = prov_type.__wrapped_type__

            for _, operator in selector.items(lambda t: issubclass(t, res_type)):
                res = (await self.setup_resource(prov_type, operator)) \
                    .map_err(lambda e: ServiceError.AutoProvision(details=e.message))
                if not res:
                    return res

        for prov_type, member in self._provider_members.items():
            if len(member.params) > 0:
                continue

            # noinspection PyTypeChecker
            match await invoke_async(member.method, self):
                case Ok(prov) | (Provider() as prov):
                    if collector.add(prov_type, prov):
                        self._provided_providers[prov_type] = prov

                case Err(e):
                    return ServiceError.Collect(details=e.message).to_result()

        await self._switch_to_active()
        return Ok(self.state)

    @override
    async def teardown(
        self, selector: ListSelector[type, Operator],
        collector: MutCollector[type[Provider], Provider],
    ) -> Result[ServiceState, ServiceError]:
        if self.state == ServiceState.Idle:
            return Ok(self.state)

        for prov_type in reversed(self._public_providers):
            res_type = prov_type.__wrapped_type__

            for _, operator in selector.items(lambda t: issubclass(t, res_type)):
                res = (await self.teardown_resource(prov_type, operator)) \
                    .map_err(lambda e: ServiceError.AutoProvision(details=e.message))
                if not res:
                    return res

        provs_to_remove = []

        for prov_type, prov in reversed(self._provided_providers.items()):
            if collector.remove(prov_type, prov):
                provs_to_remove.append(prov_type)
            else:
                # TODO logger warn
                pass

        for prov_type in provs_to_remove:
            del self._provided_providers[prov_type]

        await self._switch_to_idle()
        return Ok(self.state)

    @override
    def can_provide(self, resource_type: type) -> bool:
        if resource_type not in self._providers:
            return False
        return len(self._providers[resource_type]) > 0

    @override
    def get_resource_types(self) -> list[type]:
        return list(self._providers.keys())

    @override
    def get_provider_types(self, resource_type: type) -> list[type[Provider]]:
        if resource_type not in self._providers:
            return []
        return self._providers[resource_type].copy()

    @override
    async def setup_resource(
        self, prov_type: type[Provider], operator: Operator, *args, **kwargs,
    ) -> Result[None, ProvisionError]:
        if not (member := self._provider_members.get(prov_type)):
            return ProvisionError.NoSuchProvider(type_=prov_type.__name__).to_result()
        if operator in self._provided_resources[prov_type.__wrapped_type__]:
            return Ok(None)

        # noinspection PyTypeChecker
        match await invoke_async(member.method, self, *args, **kwargs):
            case Ok(prov) | (Provider() as prov):
                mapper = AsyncResourceMapper[prov_type](prov, operator)

                if res := (await mapper.setup()).map_err(lambda e: ProvisionError.Provision(details=e)):
                    self._provided_resources[prov_type.__wrapped_type__][operator] = mapper
                return res

            case Err(e):
                return e

    @override
    async def teardown_resource(
        self, prov_type: type[Provider], operator: Operator,
    ) -> Result[None, ProvisionError]:
        if not (mappers := self._provided_resources.get(prov_type.__wrapped_type__)):
            return Ok(None)
        if not (mapper := mappers.get(operator)):
            return Ok(None)

        if res := (await mapper.teardown()).map_err(lambda e: ProvisionError.Provision(details=e)):
            del mappers[operator]
        return res

    @override
    async def _switch_to_idle(self):
        self._state = ServiceState.Idle

    @override
    async def _switch_to_active(self, started: bool = False):
        if started:
            self._waiter.clear()
        else:
            self._waiter.set()

        self._state = ServiceState.Active(started=not self._waiter.is_set())



from typing import override

from greyhorse.result import Result, Ok
from ..abc.collectors import Collector, MutCollector
from ..abc.controllers import Controller, ControllerError
from ..abc.operators import Operator
from ..abc.providers import Provider
from ..abc.selectors import Selector


class SyncController(Controller):
    @override
    def setup(
        self, selector: Selector[type[Provider], Provider],
        collector: Collector[type, Operator],
    ) -> Result[bool, ControllerError]:
        return Ok(True)

    @override
    def teardown(
        self, selector: Selector[type[Provider], Provider],
        collector: MutCollector[type, Operator],
    ) -> Result[bool, ControllerError]:
        return Ok(True)


class AsyncController(Controller):
    @override
    async def setup(
        self, selector: Selector[type[Provider], Provider],
        collector: Collector[type, Operator],
    ) -> Result[bool, ControllerError]:
        return Ok(True)

    @override
    async def teardown(
        self, selector: Selector[type[Provider], Provider],
        collector: MutCollector[type, Operator],
    ) -> Result[bool, ControllerError]:
        return Ok(True)



from collections import defaultdict
from typing import Callable, Mapping, Type, TypeVar

from greyhorse.logging import logger
from greyhorse.result import Result
from greyhorse.utils.invoke import invoke_async, invoke_sync

MessageType = TypeVar('MessageType')
CommandHandler = Callable[[MessageType], Result]
EventHandler = Callable[[MessageType], None]


class SyncMessageBus:
    def __init__(
        self, cmd_handlers: Mapping[Type[MessageType], CommandHandler] | None = None,
        event_handlers: Mapping[Type[MessageType], list[EventHandler]] | None = None,
    ):
        self._cmd_handlers = cmd_handlers or dict()
        self._event_handlers = event_handlers or defaultdict(list)

    def add_command_handler(self, message_type: Type[MessageType], handler: CommandHandler):
        self._cmd_handlers[message_type] = handler

    def add_event_handler(self, message_type: Type[MessageType], handler: EventHandler):
        self._event_handlers[message_type].append(handler)

    def handle_command(self, message: MessageType) -> Result:
        if handler := self._cmd_handlers.get(type(message)):
            return invoke_sync(handler, message)
        raise LookupError()

    def handle_event(self, message: MessageType):
        for handler in self._event_handlers[type(message)]:
            try:
                invoke_sync(handler, message)
            except Exception as e:
                logger.exception(str(e))
                continue


class AsyncMessageBus:
    def __init__(
        self, cmd_handlers: Mapping[Type[MessageType], CommandHandler] | None = None,
        event_handlers: Mapping[Type[MessageType], list[EventHandler]] | None = None,
    ):
        self._cmd_handlers = cmd_handlers or dict()
        self._event_handlers = event_handlers or defaultdict(list)

    def add_command_handler(self, message_type: Type[MessageType], handler: CommandHandler):
        self._cmd_handlers[message_type] = handler

    def add_event_handler(self, message_type: Type[MessageType], handler: EventHandler):
        self._event_handlers[message_type].append(handler)

    async def handle_command(self, message: MessageType) -> Result:
        if handler := self._cmd_handlers.get(type(message)):
            return await invoke_async(handler, message)
        raise LookupError()

    async def handle_event(self, message: MessageType):
        for handler in self._event_handlers[type(message)]:
            try:
                await invoke_async(handler, message)
            except Exception as e:
                logger.exception(str(e))
                continue



from abc import ABC, abstractmethod


class Collector[K, T](ABC):
    @abstractmethod
    def add(self, key: K, instance: T) -> bool:
        ...


class MutCollector[K, T](Collector[K, T], ABC):
    @abstractmethod
    def remove(self, key: K, instance: T) -> bool:
        ...



from abc import ABC, abstractmethod
from typing import Awaitable, Mapping, override

from greyhorse.app.context import Context


class DataStorageEngine(ABC):
    def __init__(self, name: str):
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    @property
    @abstractmethod
    def active(self) -> bool:
        ...

    def start(self) -> Awaitable[None] | None:
        pass

    def stop(self) -> Awaitable[None] | None:
        pass

    @abstractmethod
    def get_context[T: Context](self, kind: type[T]) -> T | None:
        ...


class DataStorageEngineFactory(ABC):
    @abstractmethod
    def create_engine(self, name: str, *args, **kwargs) -> DataStorageEngine:
        ...

    @abstractmethod
    def destroy_engine(self, name: str) -> bool:
        ...

    @abstractmethod
    def get_engine_names(self) -> list[str]:
        ...

    @abstractmethod
    def get_engine(self, name: str) -> DataStorageEngine | None:
        ...

    @abstractmethod
    def get_engines(self, names: list[str] | None = None) -> Mapping[str, DataStorageEngine]:
        ...


class SimpleDataStorageFactory[T](DataStorageEngineFactory, ABC):
    def __init__(self):
        self._engines: dict[str, T] = {}

    @override
    def destroy_engine(self, name: str) -> bool:
        return self._engines.pop(name, None) is not None

    @override
    def get_engine_names(self) -> list[str]:
        return list(self._engines.keys())

    @override
    def get_engine(self, name: str) -> T | None:
        return self._engines.get(name)

    @override
    def get_engines(self, names: list[str] | None = None) -> Mapping[str, T]:
        engines = dict()

        if not names:
            names = self.get_engine_names()

        for name in names:
            if engine := self.get_engine(name):
                engines[name] = engine

        return engines



from greyhorse.app.abc.operators import Operator
from greyhorse.app.abc.providers import Provider
from greyhorse.app.entities.components import Component
from greyhorse.app.registries import MutDictRegistry
from greyhorse.app.schemas.components import ModuleConf
from greyhorse.error import Error, ErrorCase
from greyhorse.logging import logger
from greyhorse.maybe import Maybe
from greyhorse.result import Result, Ok


class ModuleError(Error):
    namespace = 'greyhorse.app.module'

    Component = ErrorCase(
        msg='{path}: Component error in module, details: "{details}"',
        path=str, details=str,
    )


class Module:
    def __init__(
        self, path: str, conf: ModuleConf, components: list[Component],
        operator_reg: MutDictRegistry[type, Operator],
    ):
        self._conf = conf
        self._path = path

        self._operator_reg = operator_reg
        self._private_providers = MutDictRegistry[type[Provider], Provider]()
        self._public_providers = MutDictRegistry[type[Provider], Provider]()
        self._components: dict[str, Component] = {c.name: c for c in components}

    @property
    def path(self) -> str:
        return self._path

    def get_component(self, name: str) -> Maybe[Component]:
        return Maybe(self._components.get(name))

    def get_provider[P: Provider](self, type_: type[P]) -> Maybe[P]:
        return self._public_providers.get(type_)

    def add_provider(self, type_: type[Provider], provider: Provider) -> bool:
        return self._public_providers.add(type_, provider)

    def remove_provider(self, type_: type[Provider], provider: Provider) -> bool:
        return self._public_providers.remove(type_, provider)

    def setup(self) -> Result[None, ModuleError]:
        logger.info('{path}: Module setup'.format(path=self._path))

        for prov_conf in self._conf.provider_claims:
            for prov_type in prov_conf.types:
                for _, prov in self._public_providers.items(lambda t: issubclass(t, prov_type)):
                    self._private_providers.add(prov_type, prov)

        for component in self._components.values():
            if not (res := component.setup(self._private_providers, self._private_providers).map_err(
                lambda e: ModuleError.Component(path=self._path, details=e.message)
            )):
                return res

        for prov_conf in self._conf.provider_exports:
            for prov_type in prov_conf.types:
                for _, prov in self._private_providers.items(lambda t: issubclass(t, prov_type)):
                    self._public_providers.add(prov_type, prov)

        logger.info('{path}: Module setup successful'.format(path=self._path))
        return Ok(None)

    def teardown(self) -> Result[None, ModuleError]:
        logger.info('{path}: Module teardown'.format(path=self._path))

        for prov_conf in reversed(self._conf.provider_exports):
            for prov_type in prov_conf.types:
                for _, prov in self._private_providers.items(lambda t: issubclass(t, prov_type)):
                    self._public_providers.remove(prov_type, prov)

        for component in reversed(self._components.values()):
            if not (res := component.teardown(self._private_providers, self._private_providers).map_err(
                lambda e: ModuleError.Component(path=self._path, details=e.message)
            )):
                return res

        for prov_conf in reversed(self._conf.provider_claims):
            for prov_type in prov_conf.types:
                for _, prov in self._public_providers.items(lambda t: issubclass(t, prov_type)):
                    self._private_providers.remove(prov_type, prov)

        logger.info('{path}: Module teardown successful'.format(path=self._path))
        return Ok(None)



from __future__ import annotations

import asyncio
import threading
from abc import ABC, abstractmethod
from collections import defaultdict
from contextlib import AbstractAsyncContextManager, AbstractContextManager, AsyncExitStack, ExitStack
from contextvars import ContextVar
from dataclasses import dataclass, field
from typing import Any, AsyncContextManager, Awaitable, Callable, ContextManager, Mapping, override
from uuid import uuid4

from greyhorse.enum import Enum, Struct, Unit
from greyhorse.maybe import Maybe, Nothing, Just
from greyhorse.utils.invoke import get_asyncio_loop, invoke_async, invoke_sync, is_like_sync_context_manager, \
    is_like_async_context_manager
from greyhorse.utils.types import TypeWrapper

type FieldFactory[T] = (
    T | Callable[[], Awaitable[T] | T] |
    AbstractContextManager[T] | AbstractAsyncContextManager[T] |
    Callable[[], AbstractContextManager[T]] | Callable[[], AbstractAsyncContextManager[T]]
)


type ContextManagerLike[T] = (
    AbstractContextManager[T] | AbstractAsyncContextManager[T] |
    Callable[[], AbstractContextManager[T]] | Callable[[], AbstractAsyncContextManager[T]]
)


class ContextState[T](Enum):
    Idle = Unit()
    InUse = Struct(count=int, value=T)
    Applied = Struct(count=int, value=T)
    Cancelled = Struct(count=int, value=T)


@dataclass(slots=True, frozen=True)
class ContextData[T]:
    factory: Callable[[...], T]
    ident: str = field(default_factory=lambda: str(uuid4()))
    field_factories: dict[str, FieldFactory[Any]] = field(default_factory=dict)
    finalizers: list[Callable[[], Awaitable[None] | None]] = field(default_factory=list)


class InvalidContextState(RuntimeError):
    pass


class Context(object):
    __slots__ = ('_state', '_data', '_parent', '_prev')

    def __init__(
        self, factory: Callable[[...], Any],
        fields: Mapping[str, FieldFactory[Any]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
    ):
        self._state = ContextState[Any].Idle
        self._data = ContextData(
            factory=factory, field_factories=fields or {},
            finalizers=finalizers or [],
        )
        self._parent: Maybe[Context] = Nothing
        self._prev: Maybe[Context] = Nothing

    @property
    def ident(self) -> str:
        return self._data.ident

    def __enter__(self) -> Any:
        raise NotImplementedError

    def __exit__(self, exc_type, exc_value, traceback):
        raise NotImplementedError

    async def __aenter__(self) -> Any:
        raise NotImplementedError

    async def __aexit__(self, exc_type, exc_value, traceback):
        raise NotImplementedError

    def children(self) -> list[Context]:
        raise NotImplementedError


class MutContext(Context, ABC):
    @abstractmethod
    def apply(self) -> Awaitable[None] | None:
        ...

    @abstractmethod
    def cancel(self) -> Awaitable[None] | None:
        ...

    @abstractmethod
    def mutable_children(self) -> list[MutContext]:
        ...


class _ContextStorage(threading.local):
    __slots__ = ('_storage',)

    def __init__(self):
        self._storage: dict[type, list[Context]] = defaultdict(list)

    def add(self, kind: type, instance: Context):
        self._storage[kind].append(instance)

    def remove(self, kind: type, instance: Context):
        self._storage[kind].remove(instance)

    def get_last(self, kind: type) -> Maybe[Context]:
        if objects := self._storage.get(kind, []):
            return Just(objects[0])
        return Nothing


_current_context: ContextVar[Context] = ContextVar('_ctx')
_context_storage = _ContextStorage()


def current_context(kind: type | None = None) -> Maybe[Context]:
    if kind is None:
        return Maybe(_current_context.get(None))
    else:
        return _context_storage.get_last(kind)


def current_scope_id(kind: type | None = None) -> str:
    if kind is None:
        if ctx := _current_context.get(None):
            return ctx.ident
    else:
        if ctx := _context_storage.get_last(kind):
            return ctx.unwrap().ident

    if loop := get_asyncio_loop():
        return f'asyncio:{id(asyncio.current_task(loop))}'
    else:
        return f'thread:{threading.current_thread().ident}'


class SyncContext[T](Context, TypeWrapper[T], ContextManager):
    __slots__ = (
        '_sync_stack', '_context_managers', '_children', '_lock',
    )

    def __init__(
        self, factory: Callable[[...], T],
        fields: dict[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        contexts: list[ContextManagerLike[T]] | None = None,
    ):
        fields = fields.copy() if fields else {}
        children: list[SyncContext] = []
        context_managers: list[tuple[ContextManagerLike, str | None]] = []

        if contexts:
            for ctx in contexts:
                if is_like_sync_context_manager(ctx):
                    context_managers.append((ctx, None))

        names_to_remove = set()

        for name, value in fields.items():
            if isinstance(value, SyncContext):
                children.append(value)
                names_to_remove.add(name)
            if is_like_sync_context_manager(value):
                context_managers.append((value, name))
                names_to_remove.add(name)
            elif is_like_async_context_manager(value):
                names_to_remove.add(name)

        for name in names_to_remove:
            fields.pop(name)

        super().__init__(factory, fields, finalizers)
        self._sync_stack = ExitStack()
        self._context_managers = context_managers
        self._children = children
        self._lock = threading.Lock()

    @override
    def children(self) -> list[Context]:
        return self._children.copy()

    def _switch_to_use(self):
        self._sync_stack.__enter__()
        kwargs: dict[str, Any] = {}

        for ctx, field in self._context_managers:
            if callable(ctx):
                ctx = ctx()
            if value := self._sync_stack.enter_context(ctx):
                if field is not None:
                    kwargs[field] = value

        for name, value in self._data.field_factories.items():
            if callable(value):
                value = invoke_sync(value)
            kwargs[name] = value

        instance = self._create(**kwargs)
        self._prev = Maybe(_current_context.get(None))
        _current_context.set(self)
        self._parent = _context_storage.get_last(type(instance))
        _context_storage.add(type(instance), self)

        return instance

    def _switch_to_idle(self, instance: T, exc_type, exc_value, traceback):
        _context_storage.remove(type(instance), self)

        try:
            self._destroy(instance)
        except Exception:
            pass

        try:
            self._sync_stack.__exit__(exc_type, exc_value, traceback)
        except Exception:
            pass

        for finalizer in self._data.finalizers:
            try:
                invoke_sync(finalizer)
            except Exception:
                pass

        _current_context.set(self._prev.unwrap_or_none())
        self._prev = self._parent = Nothing

    def _create(self, **kwargs) -> T:
        return self._data.factory(**kwargs)

    def _destroy(self, instance: T):
        del instance

    def _enter(self, instance: T):
        pass

    def _exit(self, instance: T, exc_type, exc_value, traceback):
        pass

    def _nested_enter(self, instance: T):
        pass

    def _nested_exit(self, instance: T, exc_type, exc_value, traceback):
        pass

    @override
    def __enter__(self) -> T:
        with self._lock:
            match self._state:
                case ContextState.Idle:
                    value = self._switch_to_use()
                    self._enter(value)
                    self._state = ContextState[type(value)].InUse(count=1, value=value)
                    return value

                case ContextState.InUse(count, value):
                    self._nested_enter(value)
                    self._state = ContextState[type(value)].InUse(count=count + 1, value=value)
                    return value

                case ContextState.Applied(count, value):
                    self._nested_enter(value)
                    self._state = ContextState[type(value)].InUse(count=count + 1, value=value)
                    return value

                case ContextState.Cancelled(count, value):
                    self._nested_enter(value)
                    self._state = ContextState[type(value)].InUse(count=count + 1, value=value)
                    return value

    @override
    def __exit__(self, exc_type, exc_value, traceback):
        with self._lock:
            match self._state:
                case ContextState.Idle:
                    raise InvalidContextState('Context exit on idle state')

                case ContextState.InUse(count, value) \
                        | ContextState.Applied(count, value) \
                        | ContextState.Cancelled(count, value):
                    if count > 1:
                        self._nested_exit(value, exc_type, exc_value, traceback)
                        self._state = self._state.__class__(count=count - 1, value=value)
                    else:
                        self._exit(value, exc_type, exc_value, traceback)
                        self._switch_to_idle(value, exc_type, exc_value, traceback)
                        self._state = ContextState[type(value)].Idle


class SyncMutContext[T](SyncContext[T], MutContext):
    __slots__ = (
        '_mut_children', '_force_rollback', '_auto_apply',
    )

    def __init__(
        self, factory: Callable[[...], T],
        fields: dict[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        contexts: list[ContextManagerLike[T]] | None = None,
        force_rollback: bool = False, auto_apply: bool = False,
    ):
        super().__init__(factory, fields, finalizers, contexts)
        self._mut_children = []
        self._force_rollback = force_rollback
        self._auto_apply = auto_apply

        for child in self._children:
            if isinstance(child, SyncMutContext):
                self._mut_children.append(child)

    @override
    def mutable_children(self) -> list[MutContext]:
        return self._mut_children.copy()

    def _apply(self, instance: T):
        pass

    def _cancel(self, instance: T):
        pass

    @override
    def apply(self) -> None:
        with self._lock:
            match self._state:
                case ContextState.Idle:
                    raise InvalidContextState('MutContext apply on idle state')

                case ContextState.InUse(count, value):
                    for child in self._mut_children:
                        child.apply()
                    self._apply(value)
                    self._state = ContextState[type(value)].Applied(count=count, value=value)

                case ContextState.Applied(_, _):
                    pass

                case ContextState.Cancelled(_, _):
                    raise InvalidContextState('MutContext apply on cancelled state')

    @override
    def cancel(self) -> None:
        with self._lock:
            match self._state:
                case ContextState.Idle:
                    raise InvalidContextState('MutContext cancel on idle state')

                case ContextState.InUse(count, value):
                    for child in self._mut_children:
                        child.cancel()
                    self._cancel(value)
                    self._state = ContextState[type(value)].Cancelled(count=count, value=value)

                case ContextState.Applied(_, _):
                    raise InvalidContextState('MutContext cancel on applied state')

                case ContextState.Cancelled(_, _):
                    pass

    @override
    def _exit(self, instance: T, exc_type, exc_value, traceback):
        if self._force_rollback or exc_type is not None:
            self.cancel()
        elif self._auto_apply:
            self.apply()


class AsyncContext[T](Context, TypeWrapper[T], AsyncContextManager):
    __slots__ = (
        '_sync_stack', '_async_stack',
        '_sync_sub_contexts', '_async_sub_contexts',
        '_sync_children', '_async_children', '_lock',
    )

    def __init__(
        self, factory: Callable[[...], T],
        fields: dict[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        contexts: list[ContextManagerLike[T]] | None = None,
    ):
        fields = fields.copy() if fields else {}
        sync_children: list[SyncContext] = []
        async_children: list[AsyncContext] = []
        sync_sub_contexts: list[tuple[ContextManagerLike[T], str | None]] = []
        async_sub_contexts: list[tuple[ContextManagerLike[T], str | None]] = []

        if contexts:
            for ctx in contexts:
                if is_like_async_context_manager(ctx):
                    async_sub_contexts.append((ctx, None))
                elif is_like_sync_context_manager(ctx):
                    sync_sub_contexts.append((ctx, None))

        names_to_remove = set()

        for name, value in fields.items():
            if isinstance(value, AsyncContext):
                async_children.append(value)
                names_to_remove.add(name)
            elif isinstance(value, SyncContext):
                sync_children.append(value)
                names_to_remove.add(name)
            if is_like_async_context_manager(value):
                async_sub_contexts.append((value, name))
                names_to_remove.add(name)
            elif is_like_sync_context_manager(value):
                sync_sub_contexts.append((value, name))
                names_to_remove.add(name)

        for name in names_to_remove:
            fields.pop(name)

        super().__init__(factory, fields, finalizers)
        self._sync_stack = ExitStack()
        self._async_stack = AsyncExitStack()
        self._sync_sub_contexts = sync_sub_contexts
        self._async_sub_contexts = async_sub_contexts
        self._sync_children = sync_children
        self._async_children = async_children
        self._lock = asyncio.Lock()

    @override
    def children(self) -> list[Context]:
        return self._sync_children.copy() + self._async_children.copy()

    async def _switch_to_use(self):
        self._sync_stack.__enter__()
        await self._async_stack.__aenter__()
        kwargs: dict[str, Any] = {}

        for ctx, field in self._sync_sub_contexts:
            if callable(ctx):
                ctx = ctx()
            if value := self._sync_stack.enter_context(ctx):
                if field is not None:
                    kwargs[field] = value

        for ctx, field in self._async_sub_contexts:
            if callable(ctx):
                ctx = ctx()
            if value := await self._async_stack.enter_async_context(ctx):
                if field is not None:
                    kwargs[field] = value

        for name, value in self._data.field_factories.items():
            if callable(value):
                value = await invoke_async(value)
            kwargs[name] = value

        instance = await self._create(**kwargs)
        self._prev = Maybe(_current_context.get(None))
        _current_context.set(self)
        self._parent = _context_storage.get_last(type(instance))
        _context_storage.add(type(instance), self)

        return instance

    async def _switch_to_idle(self, instance: T, exc_type, exc_value, traceback):
        _context_storage.remove(type(instance), self)

        try:
            await self._destroy(instance)
        except Exception:
            pass

        try:
            await self._async_stack.__aexit__(exc_type, exc_value, traceback)
        except Exception:
            pass

        try:
            self._sync_stack.__exit__(exc_type, exc_value, traceback)
        except Exception:
            pass

        for finalizer in self._data.finalizers:
            try:
                await invoke_async(finalizer)
            except Exception:
                pass

        _current_context.set(self._prev.unwrap_or_none())
        self._prev = self._parent = Nothing

    async def _create(self, **kwargs) -> T:
        return self._data.factory(**kwargs)

    async def _destroy(self, instance: T):
        del instance

    async def _enter(self, instance: T):
        pass

    async def _exit(self, instance: T, exc_type, exc_value, traceback):
        pass

    async def _nested_enter(self, instance: T):
        pass

    async def _nested_exit(self, instance: T, exc_type, exc_value, traceback):
        pass

    @override
    async def __aenter__(self) -> T:
        async with self._lock:
            match self._state:
                case ContextState.Idle:
                    value = await self._switch_to_use()
                    await self._enter(value)
                    self._state = ContextState[type(value)].InUse(count=1, value=value)
                    return value

                case ContextState.InUse(count, value):
                    await self._nested_enter(value)
                    self._state = ContextState[type(value)].InUse(count=count + 1, value=value)
                    return value

                case ContextState.Applied(count, value):
                    await self._nested_enter(value)
                    self._state = ContextState[type(value)].InUse(count=count + 1, value=value)
                    return value

                case ContextState.Cancelled(count, value):
                    await self._nested_enter(value)
                    self._state = ContextState[type(value)].InUse(count=count + 1, value=value)
                    return value

    @override
    async def __aexit__(self, exc_type, exc_value, traceback):
        async with self._lock:
            match self._state:
                case ContextState.Idle:
                    raise InvalidContextState('Context exit on idle state')

                case ContextState.InUse(count, value) \
                        | ContextState.Applied(count, value) \
                        | ContextState.Cancelled(count, value):
                    if count > 1:
                        await self._nested_exit(value, exc_type, exc_value, traceback)
                        self._state = self._state.__class__(count=count - 1, value=value)
                    else:
                        await self._exit(value, exc_type, exc_value, traceback)
                        await self._switch_to_idle(value, exc_type, exc_value, traceback)
                        self._state = ContextState[type(value)].Idle


class AsyncMutContext[T](AsyncContext[T], MutContext):
    __slots__ = (
        '_sync_mut_children', '_async_mut_children', '_force_rollback', '_auto_apply',
    )

    def __init__(
        self, factory: Callable[[...], T],
        fields: dict[str, FieldFactory[T]] | None = None,
        finalizers: list[Callable[[], Awaitable[None] | None]] | None = None,
        contexts: list[ContextManagerLike[T]] | None = None,
        force_rollback: bool = False, auto_apply: bool = False,
    ):
        super().__init__(factory, fields, finalizers, contexts)
        self._sync_mut_children = []
        self._async_mut_children = []
        self._force_rollback = force_rollback
        self._auto_apply = auto_apply

        for child in self._sync_children:
            if isinstance(child, SyncMutContext):
                self._sync_mut_children.append(child)

        for child in self._async_children:
            if isinstance(child, AsyncMutContext):
                self._async_mut_children.append(child)

    @override
    def mutable_children(self) -> list[MutContext]:
        return self._sync_mut_children.copy() + self._async_mut_children.copy()

    async def _apply(self, instance: T):
        pass

    async def _cancel(self, instance: T):
        pass

    @override
    async def apply(self) -> None:
        async with self._lock:
            match self._state:
                case ContextState.Idle:
                    raise InvalidContextState('MutContext apply on idle state')

                case ContextState.InUse(count, value):
                    for child in self._sync_mut_children:
                        child.apply()
                    for child in self._async_mut_children:
                        await child.apply()
                    await self._apply(value)
                    self._state = ContextState[type(value)].Applied(count=count, value=value)

                case ContextState.Applied(_, _):
                    pass

                case ContextState.Cancelled(_, _):
                    raise InvalidContextState('MutContext apply on cancelled state')

    @override
    async def cancel(self) -> None:
        async with self._lock:
            match self._state:
                case ContextState.Idle:
                    raise InvalidContextState('MutContext cancel on idle state')

                case ContextState.InUse(count, value):
                    for child in self._sync_mut_children:
                        child.cancel()
                    for child in self._async_mut_children:
                        await child.cancel()
                    await self._cancel(value)
                    self._state = ContextState[type(value)].Cancelled(count=count, value=value)

                case ContextState.Applied(_, _):
                    raise InvalidContextState('MutContext cancel on applied state')

                case ContextState.Cancelled(_, _):
                    pass

    @override
    async def _exit(self, instance: T, exc_type, exc_value, traceback):
        if self._force_rollback or exc_type is not None:
            await self.cancel()
        elif self._auto_apply:
            await self.apply()


class ContextBuilder[T](TypeWrapper[T]):
    def __init__(self, factory: Callable[[...], T], **kwargs: dict[str, Any]):
        self._factory = factory
        self._fields = {}
        self._finalizers = []
        self._contexts = []
        self._kwargs = kwargs

    def add_param(self, name: str, value: FieldFactory[T]):
        self._fields[name] = value

    def add_context(self, context: ContextManagerLike[T]):
        self._contexts.append(context)

    def add_finalizer(self, finalizer: Callable[[], Awaitable[None] | None]):
        self._finalizers.append(finalizer)

    def build(self):
        return self.wrapped_type(
            factory=self._factory, fields=self._fields,
            finalizers=self._finalizers, contexts=self._contexts,
            **self._kwargs,
        )

    def __class_getitem__[C: SyncContext | AsyncContext | SyncMutContext | AsyncMutContext](
        cls, args: tuple[type[C], type[T]]
    ) -> ContextBuilder:
        class_, type_ = args
        return super(ContextBuilder, cls).__class_getitem__(class_[type_])



from abc import ABC, abstractmethod
from typing import Awaitable, override, TypeVar

from greyhorse.app.abc.operators import Operator
from greyhorse.app.abc.providers import Provider, SharedProvider, MutProvider, FactoryProvider, ForwardProvider
from greyhorse.result import Result, Ok, Err
from greyhorse.utils.types import TypeWrapper


class ResourceMapper[T](TypeWrapper[T], ABC):
    __slots__ = ('_provider', '_operator', '_methods')

    def __init__(self, provider: Provider[T], operator: Operator[T]):
        self._provider = provider
        self._operator = operator

    def __class_getitem__(cls, provider_type: type[Provider]):
        if isinstance(provider_type, TypeVar):
            # noinspection PyUnresolvedReferences
            return super(TypeWrapper, cls).__class_getitem__(provider_type)

        class_ = super(ResourceMapper, cls).__class_getitem__(provider_type.__wrapped_type__)
        setattr(class_, '_methods', cls._get_provider_methods(provider_type))
        return class_

    @staticmethod
    def _get_provider_methods(provider_type: type[Provider]):
        if issubclass(provider_type, SharedProvider):
            return 'borrow', 'reclaim'
        elif issubclass(provider_type, MutProvider):
            return 'acquire', 'release'
        elif issubclass(provider_type, FactoryProvider):
            return 'create', 'destroy'
        elif issubclass(provider_type, ForwardProvider):
            return 'take', 'drop'

    @abstractmethod
    def setup(self) -> Result[bool, str] | Awaitable[Result[bool, str]]:
        ...

    @abstractmethod
    def teardown(self) -> Result[None, str] | Awaitable[Result[None, str]]:
        ...


class SyncResourceMapper[T](ResourceMapper[T]):
    @override
    def setup(self) -> Result[None, str]:
        init_method = getattr(self._provider, self._methods[0])
        return init_method().map_err(lambda e: e.message) \
            .map(self._operator.accept).and_then(
                lambda v: Ok(None) if v else Err(f'Could not accept resource "{self.__wrapped_type__.__name__}"')
            )

    @override
    def teardown(self) -> Result[None, str]:
        fini_method = getattr(self._provider, self._methods[1])
        return self._operator.revoke().map(fini_method) \
            .ok_or(f'Could not revoke resource "{self.__wrapped_type__.__name__}"')


class AsyncResourceMapper[T](ResourceMapper[T]):
    @override
    async def setup(self) -> Result[None, str]:
        init_method = getattr(self._provider, self._methods[0])
        return await (
            await (await init_method()).map_err(lambda e: e.message).map_async(self._operator.accept)
        ).and_then_async(
            lambda v: Ok(None) if v else Err(f'Could not accept resource "{self.__wrapped_type__.__name__}"')
        )

    @override
    async def teardown(self) -> Result[None, str]:
        fini_method = getattr(self._provider, self._methods[1])
        return (await (await self._operator.revoke()).map_async(fini_method)) \
            .ok_or(f'Could not revoke resource "{self.__wrapped_type__.__name__}"')



import asyncio
import signal
import threading
from asyncio import get_running_loop
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Callable, Collection, Optional, TYPE_CHECKING

from greyhorse.result import Result, Ok
from .module import Module
from ..abc.services import Service
from ...enum import Enum, Struct
from ...error import Error, ErrorCase
from ...i18n import tr
from ...logging import logger
from ...maybe import Maybe, Nothing
from ...utils.invoke import get_asyncio_loop, caller_path


class AppNotLoadedError(RuntimeError):
    pass


class ModuleInitError(Error):
    namespace = 'greyhorse.app.builder'

    Disabled = ErrorCase(msg='Module is disabled: "{path}"', path=str)
    NotLoaded = ErrorCase(msg='Module is not loaded: "{path}"', path=str)


class Application:
    def __init__(
        self, name: str, version: str = '', debug: bool = False,
    ):
        self._name = name
        self._version = version
        self._debug = debug
        self._path = self._inspect_cwd()
        self._services: dict[str, Service] = {}
        self._module: Maybe[Module] = Nothing

    @staticmethod
    def _inspect_cwd():
        import inspect

        for frame in reversed(inspect.stack()):
            path = Path(frame.filename).absolute()

            while path.parent != path:
                path = path.parent
                pyproject_toml_path = path / 'pyproject.toml'
                if pyproject_toml_path.exists():
                    return path

    @property
    def name(self) -> str:
        return self._name

    @property
    def version(self) -> str:
        return self._version

    @property
    def debug(self) -> bool:
        return self._debug

    def get_cwd(self) -> Path:
        return self._path.absolute()

    def register_service(self, name: str, instance: Service):
        self._services[name] = instance

    def unregister_service(self, name: str):
        self._services.pop(name, None)

    def get_module(self, name: str) -> Module | None:
        if value := self._modules.get(name):
            return value[0]
        return None

    def create(self) -> list[ModuleErrorsItem]:
        if self._module:
            return self._module.create()
        raise AppNotLoadedError()

    def destroy(self) -> list[ModuleErrorsItem]:
        if self._module:
            return self._module.destroy()
        raise AppNotLoadedError()

    def start(self):
        if self._module:
            return self._module.start()
        raise AppNotLoadedError()

    def stop(self):
        if self._module:
            return self._module.stop()
        raise AppNotLoadedError()

    def run_sync(self, callback: Callable[[], None] | None = None):
        sync_events: list[threading.Event] = []
        async_events: list[asyncio.Event] = []

        for name, srv in self._services.items():
            match srv.wait():
                case threading.Event() as event:
                    sync_events.append(event)
                case asyncio.Event() as event:
                    async_events.append(event)

        all_events = sync_events + async_events

        async def waiter():
            async with asyncio.TaskGroup() as tg:
                for e in async_events:
                    tg.create_task(e.wait())

        logger.info(tr('app.application.run-sync-start').format(name=self.name))

        while not all([e.is_set() for e in all_events]):
            if async_events:
                loop = get_running_loop()
                loop.run_until_complete(waiter())

            sync_events_bools = [e.wait(0.1) for e in sync_events]

            if callback and not all(sync_events_bools):
                callback()

        logger.info(tr('app.application.run-sync-stop').format(name=self.name))

    async def run_async(self):
        sync_events: list[threading.Event] = []
        async_events: list[asyncio.Event] = []

        for name, srv in self._services.items():
            match srv.wait():
                case threading.Event() as event:
                    sync_events.append(event)
                case asyncio.Event() as event:
                    async_events.append(event)

        all_events = sync_events + async_events

        logger.info(tr('app.application.run-async-start').format(name=self.name))

        while not all([e.is_set() for e in all_events]):
            async with asyncio.TaskGroup() as tg:
                for e in sync_events:
                    tg.create_task(asyncio.to_thread(e.wait, 0.1))
                for e in async_events:
                    tg.create_task(e.wait())

        logger.info(tr('app.application.run-async-stop').format(name=self.name))

    @contextmanager
    def graceful_exit(self, signals: Collection[int] = (signal.SIGINT, signal.SIGTERM)):
        signals = set(signals)
        flag: list[bool] = []

        if loop := get_asyncio_loop():
            for sig_num in signals:
                loop.add_signal_handler(sig_num, self._exit_handler, sig_num, flag)
            try:
                yield
            finally:
                for sig_num in signals:
                    loop.remove_signal_handler(sig_num)
        else:
            original_handlers = []

            for sig_num in signals:
                original_handlers.append(signal.getsignal(sig_num))
                signal.signal(sig_num, self._exit_handler)
            try:
                yield
            finally:
                for sig_num, handler in zip(signals, original_handlers):
                    signal.signal(sig_num, handler)

    def _exit_handler(self, sig_num: 'signal.Signals', flag: list[bool], *_) -> None:
        if flag:
            self._second_exit_stage(sig_num)
        else:
            self._first_exit_stage(sig_num)
            flag.append(True)

    def _first_exit_stage(self, sig_num: 'signal.Signals'):
        fail = False

        try:
            self.stop()
        except RuntimeError:
            fail = True

        if fail:
            self._second_exit_stage(sig_num)

    def _second_exit_stage(self, sig_num: 'signal.Signals'):
        raise SystemExit(128 + sig_num)



import types
from typing import TypeVar

from .strings import capitalize

_TYPES_CACHE = {}


class TypeWrapper[T]:
    __wrapped_type__ = type

    @property
    def wrapped_type(self) -> type:
        return self.__wrapped_type__

    @classmethod
    def __generate_typename__(cls, type_: type, include_base_name: bool = True) -> str:
        if isinstance(type_, TypeVar):
            type_name = '~'
        elif hasattr(type_, '__args__'):
            type_args = [cls.__generate_typename__(a, False) for a in type_.__args__]
            type_name = f'{''.join(type_args)}{capitalize(type_.__name__)}'
        else:
            type_name = f'{capitalize(type_.__name__)}'

        if include_base_name:
            type_name = type_name + cls.__name__
        return type_name

    def __init_subclass__(cls, **kwargs):
        if spec := kwargs.pop('spec', None):
            if not isinstance(spec, list):
                spec = [spec]

            for s in spec:
                base = cls.__base__
                while '~' == base.__name__[0]:
                    base = base.__base__
                base = base[s]
                type_name = cls.__generate_typename__(base, False)
                _TYPES_CACHE[type_name] = cls

        super().__init_subclass__(**kwargs)

    def __class_getitem__(cls, type_: type[T]):
        if isinstance(type_, TypeVar):
            # noinspection PyUnresolvedReferences
            return super(TypeWrapper, cls).__class_getitem__(type_)

        type_name = cls.__generate_typename__(type_)

        if class_ := _TYPES_CACHE.get(type_name):
            return class_

        base_type_name = '~' + cls.__generate_typename__(type_.__base__)
        base_class = _TYPES_CACHE.get(base_type_name, cls)

        if hasattr(cls, '__slots__'):
            cls.__slots__ = (*cls.__slots__, '__wrapped_type__')
            attrs = {k: v for k, v in cls.__dict__.items() if k in cls.__slots__}
            attrs['__wrapped_type__'] = type_
            attrs['__module__'] = type_.__module__
            class_ = types.new_class(type_name, (base_class,), exec_body=lambda d: d.update(attrs))

        else:
            attrs = dict(cls.__dict__)
            attrs.pop('__class_getitem__', None)
            attrs['__wrapped_type__'] = type_
            attrs['__module__'] = type_.__module__

            class_ = types.new_class(type_name, (base_class,), exec_body=lambda d: d.update(attrs))

        _TYPES_CACHE[type_name] = class_
        return class_



from typing import Any

from pydantic import BaseModel, Field, PrivateAttr

from greyhorse.app.abc.controllers import ControllerFactories
from greyhorse.app.abc.operators import Operator
from greyhorse.app.abc.providers import Provider
from greyhorse.app.abc.services import ServiceFactories
from greyhorse.app.schemas.elements import CtrlConf, SvcConf
from greyhorse.utils.invoke import caller_path


class ProvidersConf(BaseModel, frozen=True):
    resource: type
    types: list[type[Provider]] = Field(default_factory=list)


class OperatorsConf(BaseModel, frozen=True):
    resource: type
    types: list[type[Operator]] = Field(default_factory=list)


class BasicComponentConf(BaseModel):
    name: str = Field(frozen=True)
    enabled: bool = Field(default=True)

    provider_grants: list[ProvidersConf] = Field(default_factory=list)
    provider_imports: list[ProvidersConf] = Field(default_factory=list)


class ModuleConf(BaseModel):
    enabled: bool = Field(default=True)

    provider_claims: list[ProvidersConf] = Field(default_factory=list)
    provider_exports: list[ProvidersConf] = Field(default_factory=list)

    components: list[BasicComponentConf] = Field(default_factory=list)


class ComponentConf(BasicComponentConf):
    controllers: list[CtrlConf] = Field(default_factory=list)
    services: list[SvcConf] = Field(default_factory=list)

    controller_factories: ControllerFactories = Field(default_factory=dict)
    service_factories: ServiceFactories = Field(default_factory=dict)


class ModuleComponentConf(BasicComponentConf):
    path: str = Field(frozen=True)
    args: dict[str, Any] = Field(default_factory=dict, frozen=True)
    _init_path: list[str] = PrivateAttr(default_factory=lambda: caller_path(5))
    _conf: ModuleConf | None = PrivateAttr(default=None)



from abc import ABC, abstractmethod
from typing import Awaitable, Callable

from greyhorse.error import Error, ErrorCase
from greyhorse.result import Result
from .collectors import Collector, MutCollector
from .operators import Operator
from .providers import Provider
from .selectors import Selector


class ControllerError(Error):
    namespace = 'greyhorse.app'

    Factory = ErrorCase(msg='Controller factory error: "{details}"', details=str)
    Deps = ErrorCase(msg='Dependency error occurred: "{details}"', details=str)


type ControllerFactoryFn = Callable[[...], Controller | Result[Controller, ControllerError]]
type ControllerFactories = dict[type[Controller], ControllerFactoryFn]


class Controller(ABC):
    @abstractmethod
    def setup(
        self, selector: Selector[type[Provider], Provider],
        collector: Collector[type, Operator],
    ) -> Result[bool, ControllerError] | Awaitable[Result[bool, ControllerError]]:
        ...

    @abstractmethod
    def teardown(
        self, selector: Selector[type[Provider], Provider],
        collector: MutCollector[type, Operator],
    ) -> Result[bool, ControllerError] | Awaitable[Result[bool, ControllerError]]:
        ...



from typing import Any, Mapping, Protocol


class Serializer(Protocol):
    def serialize(self, data: Any | None = None) -> bytes:
        ...


class Deserializer(Protocol):
    def deserialize(self, data: bytes) -> Any | None:
        ...


class ModelSerializer(Protocol):
    serializer: Serializer

    def serialize(self, only_fields: list[str] = None) -> bytes:
        return self.serializer.serialize(self.get_serializable_values(only_fields))

    def get_serializable_values(self, only_fields: list[str] = None) -> Mapping[str, Any] | None:
        ...



from __future__ import annotations

from dataclasses import fields as dataclass_fields
from typing import ClassVar, Final, TYPE_CHECKING, Any, Self

from .enum import Struct, Enum
from .i18n import StaticTranslator

if TYPE_CHECKING:
    from .result import Result


class Error(Enum):
    _tr: ClassVar[StaticTranslator] = None
    namespace: ClassVar[str] = ''
    code: Final[str] = ''
    msg: Final[str] = ''

    @property
    def message(self) -> str:
        values = {}

        # noinspection PyDataclass
        for field in dataclass_fields(self):
            if not field.repr or not field.init:
                continue

            values[field.name] = getattr(self, field.name, '')

        if self.code and self._tr is not None:
            msg = self._tr('.'.join([self.namespace, self.code]), default=self.msg)
        else:
            msg = self.msg

        msg = msg.format(**values)
        return msg

    def __init_subclass__(cls, **kwargs):
        if 'tr' in kwargs:
            assert isinstance(kwargs['tr'], StaticTranslator)
            cls._tr = kwargs.pop('tr')
        return super().__init_subclass__(**kwargs)

    def to_result(self) -> 'Result[Any, Self]':
        from .result import Err

        return Err(self)


class ErrorCase(Struct):
    __slots__ = ('msg', 'code')

    def _repr_fn(self, instance):
        return f'{self._base.__name__}:{self._name}(\"{instance.message}\")'



import asyncio
import threading
from abc import ABC, abstractmethod
from typing import Awaitable, Callable

from greyhorse.enum import Enum, Unit, Tuple, Struct
from greyhorse.error import Error, ErrorCase
from greyhorse.result import Result
from .collectors import Collector, MutCollector
from .operators import Operator
from .providers import Provider
from .selectors import ListSelector


class ServiceState(Enum):
    Idle = Unit()
    Active = Struct(started=bool)


class ServiceWaiter(Enum):
    Sync = Tuple(threading.Event)
    Async = Tuple(asyncio.Event)


class ServiceError(Error):
    namespace = 'greyhorse.app'

    Factory = ErrorCase(msg='Service factory error: "{details}"', details=str)
    AutoProvision = ErrorCase(msg='Service error occurred during setup resources: "{details}"', details=str)
    Collect = ErrorCase(msg='Service error occurred during collect providers: "{details}"', details=str)


type ServiceFactoryFn = Callable[[...], Service | Result[Service, ServiceError]]
type ServiceFactories = dict[type[Service], ServiceFactoryFn]


class ProvisionError(Error):
    namespace = 'greyhorse.app'

    WrongState = ErrorCase(
        msg='Cannot create provider "{type_}" because service is in wrong state: "{state}"',
        name=str, state=str,
    )

    InsufficientDeps = ErrorCase(
        msg='Cannot create provider "{type_}" because dependencies are not enough to satisfy',
        name=str,
    )

    NoSuchProvider = ErrorCase(msg='No such provider: "{type_}"', type_=str)
    Provision = ErrorCase(msg='Provision error occurred: "{details}"', details=str)


class Service(ABC):
    def __init__(self):
        self._state = ServiceState.Idle

    @property
    def state(self) -> ServiceState:
        return self._state

    @property
    @abstractmethod
    def waiter(self) -> ServiceWaiter:
        ...

    @abstractmethod
    def setup(
        self, selector: ListSelector[type, Operator],
        collector: Collector[type[Provider], Provider],
    ) -> Result[ServiceState, ServiceError] | Awaitable[Result[ServiceState, ServiceError]]:
        ...

    @abstractmethod
    def teardown(
        self, selector: ListSelector[type, Operator],
        collector: MutCollector[type[Provider], Provider],
    ) -> Result[ServiceState, ServiceError] | Awaitable[Result[ServiceState, ServiceError]]:
        ...

    @abstractmethod
    def can_provide(self, resource_type: type) -> bool:
        ...

    @abstractmethod
    def get_resource_types(self) -> list[type]:
        ...

    @abstractmethod
    def get_provider_types(self, resource_type: type) -> list[type[Provider]]:
        ...

    @abstractmethod
    def setup_resource(
        self, prov_type: type[Provider], operator: Operator, *args, **kwargs,
    ) -> Result[None, ProvisionError] | Awaitable[Result[None, ProvisionError]]:
        ...

    @abstractmethod
    def teardown_resource[T](
        self, prov_type: type[Provider], operator: Operator,
    ) -> Result[None, ProvisionError] | Awaitable[Result[None, ProvisionError]]:
        ...

    def _switch_to_idle(self):
        self._state = ServiceState.Idle

    def _switch_to_active(self, started: bool = False):
        self._state = ServiceState.Active(started=started)



from dataclasses import dataclass, field
from types import ClassMethodDescriptorType, FunctionType, GetSetDescriptorType, MappingProxyType, MemberDescriptorType, \
    MethodDescriptorType, MethodType, MethodWrapperType, WrapperDescriptorType
from typing import Any, Dict, Mapping, Protocol, Sequence, Set


@dataclass
class FieldsCache:
    priv: Set[str] = field(default_factory=set)
    pub: Set[str] = field(default_factory=set)
    ser: Set[str] = field(default_factory=set)


class ModelFieldsMixin(Protocol):
    class Meta:
        private_fields: Set[str] = set()
        public_fields: Set[str] = set()
        serializable_fields: Set[str] = set()
        non_serializable_fields: Set[str] = set()
        _fields_cache: Dict[str, FieldsCache] = dict()

    @classmethod
    def _fields_cache_key(cls):
        return '.'.join([cls.__module__, cls.__qualname__])

    # noinspection PyProtectedMember
    @classmethod
    def _calculate_fields(cls):
        private_types = (
            type, FunctionType, MethodType, MappingProxyType,
            WrapperDescriptorType, MethodWrapperType, MethodDescriptorType,
            ClassMethodDescriptorType, GetSetDescriptorType, MemberDescriptorType
        )

        private_fields, public_fields, serializable_fields = set(), set(), set()

        for attr in dir(cls):
            value = getattr(cls, attr)
            if isinstance(value, private_types) or attr[0] == '_':
                private_fields.add(attr)
            elif not isinstance(value, private_types):
                if not isinstance(value, property):
                    serializable_fields.add(attr)
                public_fields.add(attr)

        private_fields.update(cls.Meta.private_fields)
        private_fields.difference_update(cls.Meta.public_fields)
        public_fields.update(cls.Meta.public_fields)
        public_fields.difference_update(cls.Meta.private_fields)
        serializable_fields.update(cls.Meta.serializable_fields)
        serializable_fields.difference_update(cls.Meta.non_serializable_fields)
        cls.Meta._fields_cache[cls._fields_cache_key()] = \
            FieldsCache(priv=private_fields, pub=public_fields, ser=serializable_fields)

    # noinspection PyProtectedMember
    @classmethod
    def drop_fields_cache(cls):
        key = cls._fields_cache_key()
        cls.Meta._fields_cache.pop(key, None)

    # noinspection PyProtectedMember
    @classmethod
    def get_private_fields(cls) -> set[str]:
        key = cls._fields_cache_key()
        if key not in cls.Meta._fields_cache:
            cls._calculate_fields()
        return cls.Meta._fields_cache[key].priv

    # noinspection PyProtectedMember
    @classmethod
    def get_public_fields(cls) -> set[str]:
        key = cls._fields_cache_key()
        if key not in cls.Meta._fields_cache:
            cls._calculate_fields()
        return cls.Meta._fields_cache[key].pub

    @classmethod
    def get_fields(cls) -> set[str]:
        return cls.get_public_fields()

    def get_values(
        self, only_fields: Sequence[str] = None,
            exclude_fields: Sequence[str] = None) -> Mapping[str, Any]:
        public_fields = self.get_public_fields()
        only_fields = set(only_fields) if only_fields else set()
        exclude_fields = set(exclude_fields) if exclude_fields else set()

        return {
            name: getattr(self, name) for name in public_fields
            if hasattr(self, name) and (not only_fields or name in only_fields)
            and (not exclude_fields or name not in exclude_fields)
        }



from typing import Any

from pydantic import BaseModel, Field, model_validator, AliasChoices

from ..abc.controllers import Controller
from ..abc.operators import Operator
from ..abc.providers import Provider
from ..abc.services import Service


class CtrlConf(BaseModel, frozen=True):
    type_: type[Controller] = Field(validation_alias=AliasChoices('type'))
    name: str | None = None
    args: dict[str, Any] = Field(default_factory=dict)
    operators: list[type[Operator]] = Field(default_factory=list)

    @model_validator(mode='before')
    def _setup_name(self: dict[str, Any]):
        if 'name' not in self:
            self['name'] = self['type'].__name__
        return self


class SvcConf(BaseModel, frozen=True):
    type_: type[Service] = Field(validation_alias=AliasChoices('type'))
    name: str | None = None
    args: dict[str, Any] = Field(default_factory=dict)
    providers: list[type[Provider]] = Field(default_factory=list)

    @model_validator(mode='before')
    def _setup_name(self: dict[str, Any]):
        if 'name' not in self:
            self['name'] = self['type'].__name__
        return self



import inspect
from typing import Any, Callable, Mapping

TypeProviderFactory = Callable[[str, ...], Any]


class ParamsInjector:
    def __init__(self, type_providers: TypeProviderFactory | None = None):
        self._types_providers: dict[type, TypeProviderFactory | Any] = type_providers or {}

    def add_type_provider(self, t: type, p: TypeProviderFactory | Any):
        self._types_providers[t] = p

    def __call__(
        self, func: Callable[[...], ...],
        values: Mapping[str, Any] | None = None,
        types: Mapping[type, Any] | None = None,
    ):
        values = values or {}
        types = types or {}

        sig = inspect.signature(func)
        args = {}

        for k, v in sig.parameters.items():
            if 'values' == k:
                args[k] = values
            elif v.annotation in types:
                args[k] = types[v.annotation]
            elif k in values:
                args[k] = values[k]
            elif v.annotation in self._types_providers:
                args[k] = self._types_providers[v.annotation]

        args = sig.bind_partial(**args)
        args.apply_defaults()
        return args
